﻿// Copyright (c) 2015 NEC Corporation
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
// to whom the Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.CodeAnalysis.CSharp;
using NCMCSParser.Properties;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.Semantics;
using Microsoft.CodeAnalysis.Text;

namespace NCMCSParser
{
    /**
     * 意味解析(Visitor)クラス.
     * C#Parserのパーサ結果（ノード）を元に関数や条件式を解析する。
     */
    public class NCMCSVisitor : CSharpSyntaxWalker
    {
        bool flgOutsideFunction = false;
        bool flgFunctionBlock = false;
        bool flgBranchCondition = false;
        private List<String> scopenames = new List<String>();

        private Stack<TokenData> ppIfStack = new Stack<TokenData>();
        private List<string> targetCodeList = new List<string>();


        private List<int> commentLineList = null;


        public NCMCSVisitor() : base(SyntaxWalkerDepth.StructuredTrivia)
        {

        }


        /**
         * 行単位で測定情報を出力する（"--debug2"のみ有効）
         * Outputs the measurement information on a row
         * by row basis.
         */
        void debugPrintLineDetail(NCMStructData.NCMFuncData fncdata)
        {
            const string LOG_HEADER = "LINE,ELOC,CLOC,SKIP, CND,WHIL, FOR,  IF,SWCH,CASE,DFLT, AND,NEST";
            const int MAX_LENGTH = 70;
            string MAX_FORMAT = string.Format("{{0,-{0}}}", MAX_LENGTH);

            int startKey = 0;
            int endKey = 0;
            foreach (var item in fncdata.counterList)
            {
                if (startKey == 0)
                {
                    startKey = item.Key;
                }

                if (endKey < item.Key)
                {
                    endKey = item.Key;
                }

            }

            string line = "----";
            for (int i = 0; i < LOG_HEADER.Split(',').Length - 1; i++)
            {
                line += ",----";
            }
            string padding = (targetCodeList.Count != 0) ? " ".PadRight(MAX_LENGTH) + "," : string.Empty;


            NCMCSParser.structdata.dbugbufPrintLn(2, padding + LOG_HEADER); //* MSG_LOG_ID_CS_00000041

            string statement = string.Empty;
            for (int i = startKey; i <= endKey; i++)
            {
                if (targetCodeList.Count > 0)
                {

                    if (targetCodeList.Count > i - 1)
                    {
                        statement = string.Format(MAX_FORMAT, targetCodeList[i - 1]);
                    }
                    else
                    {
                        statement = string.Empty;
                    }

                    if (statement.Length > MAX_LENGTH)
                    {
                        statement = statement.Substring(0, MAX_LENGTH);
                    }

                    statement += ",";
                }

                if (fncdata.counterList != null && fncdata.counterList.ContainsKey(i))
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, "" +
                        statement +
                        string.Format("{0,4},", i) + fncdata.counterList[i]); //* MSG_LOG_ID_CS_00000042
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, "" +
                        statement +
                        string.Format("{0,4},", i)); //* MSG_LOG_ID_CS_00000043
                }
            }

            NCMCSParser.structdata.dbugbufPrintLn(2, padding + line); //* MSG_LOG_ID_CS_00000044
            NCMCSParser.structdata.dbugbufPrintLn(2,
                string.Format(padding + "{0,4},{1,4},{2,4},{3,4},{4,4},{5,4},{6,4},{7,4},{8,4},{9,4},{10,4},{11,4},{12,4}",
                    fncdata.BodyEndToken.Line - fncdata.FuncStartToken.Line + 1,
                    fncdata.cntEloc,
                    fncdata.cntCloc,
                    fncdata.cntSkip,
                    fncdata.cntCndOps,
                    fncdata.cntLoopsDoWhile,
                    fncdata.cntLoopsFor,
                    fncdata.cntConditionalIf,
                    fncdata.cntSelectionSwitch,
                    fncdata.cntSelectionCase,
                    fncdata.cntSelectionDefault,
                    fncdata.cntLogicalAnd,
                    fncdata.cntNests
                )); //* MSG_LOG_ID_CS_00000045
            NCMCSParser.structdata.dbugbufPrintLn(2, ""); //* MSG_LOG_ID_CS_00000046
        }

        /**
         * 解析対象となる行がDictionaryに存在しない場合は追加する
         * Adds the line to be analyzed if it is not present
         * in the Dictionary.
         */
        void addTargetLine(ref Dictionary<int, NCMStructData.NCMFuncData.NCMLineData> dict, int line)
        {
            if (dict == null || line < 0)
            {
                return;
            }


            if (!dict.ContainsKey(line))
            {
                NCMStructData.NCMFuncData.NCMLineData nc = new NCMStructData.NCMFuncData.NCMLineData();
                dict.Add(line, nc);
            }
        }


        /**
         * Ananlyses a given token
         * and updates concerned metric
         * value.
         */
        void CountToken(TokenData td)
        {
            NCMStructData.NCMFuncData fncdata = null;
            fncdata = NCMCSParser.structdata.functionGetData();

            if (fncdata == null || fncdata.counterList == null)
            {
                return;
            }

            if (td == null || NCMCSParser.structdata == null)
            {
                return;
            }

            // A column position less than 0 is invalid.
            if (td.StartLocation == null || td.EndLocation == null ||
               td.StartLocation.Character < 0 || td.EndLocation.Character < 0)
            {
                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000047_01 + td.ToString()); //* MSG_LOG_ID_CS_00000047
                return;
            }


            string[] rows = new string[0];
            int rows_count = 0;
            if (td.Token != null && td.Token.StartsWith("@"))
            {
                rows = td.Token.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
                rows_count = rows.Length - 1;
            }
            int startLine = td.StartLocation.Line;

            /* A SingleLineDocumentationCommentTrivia has one extra line in the
             * AST than in the actual source. Hence, to avoid metric computation
             * for the extra line, reduce the endline by one.
             */
            int endLine = (td.RoleName == SyntaxKind.SingleLineDocumentationCommentTrivia.ToString()) ?
                          td.EndLocation.Line + rows_count - 1 : td.EndLocation.Line + rows_count;

            /* If Token/Trivia is not inside a function
             * definition, then metric values will not
             * be updated.
             */
            if (startLine < fncdata.FuncStartToken.Line)
            {
                return;
            }
            
            /* Process the token from its start to end position. */			
            for (int currentLine = startLine; currentLine <= endLine; currentLine++)
            {
                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000048_01 + td.ToString()); //* MSG_LOG_ID_CS_00000048


                addTargetLine(ref fncdata.counterList, currentLine);


                switch (td.kind)
                {
                    case SyntaxKind.SingleLineCommentTrivia:
                    case SyntaxKind.MultiLineCommentTrivia:
                    case SyntaxKind.DocumentationCommentExteriorTrivia:
                    case SyntaxKind.SingleLineDocumentationCommentTrivia:
                    case SyntaxKind.MultiLineDocumentationCommentTrivia:

                        fncdata.counterList[currentLine].cntCloc++;
                        break;
                    case SyntaxKind.IfDirectiveTrivia:
                    case SyntaxKind.ElifDirectiveTrivia:
                    case SyntaxKind.ElseDirectiveTrivia:
                    case SyntaxKind.EndIfDirectiveTrivia:
                    case SyntaxKind.RegionDirectiveTrivia:
                    case SyntaxKind.EndRegionDirectiveTrivia:
                    case SyntaxKind.DefineDirectiveTrivia:
                    case SyntaxKind.UndefDirectiveTrivia:
                    case SyntaxKind.ErrorDirectiveTrivia:
                    case SyntaxKind.WarningDirectiveTrivia:
                    case SyntaxKind.LineDirectiveTrivia:
                    case SyntaxKind.PragmaWarningDirectiveTrivia:
                    case SyntaxKind.PragmaChecksumDirectiveTrivia:
                        if (fncdata.counterList[currentLine].isInactiveCode)
                        {
                            fncdata.counterList[currentLine].isInactiveCode = false;
                        }
                        /* Check the PpTake field of the token. Mark the
                         * corresponding line as inactive if PpTake is false,
                         * else increment ELOC for that particular line.
                         */
                        if (!td.PpTake)
                        {
                            markInactiveCode(td);
                        }
                        else
                        {
                            fncdata.counterList[currentLine].cntEloc++;
                        }
                        break;
                    case SyntaxKind.OpenBraceToken:

                        fncdata.cntNestsLevel++;
                        if (fncdata.cntNests < fncdata.cntNestsLevel)
                        {
                            fncdata.cntNests = fncdata.cntNestsLevel;


                            fncdata.counterList[currentLine].cntNests = (ushort)fncdata.cntNestsLevel;
                        }
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                    case SyntaxKind.CloseBraceToken:

                        fncdata.cntNestsLevel--;
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                    case SyntaxKind.CommaToken:
                    default:
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                }

            }


        }


        /**
         * プリプロセッサディレクティブ内の無効行をマークする。
         * Marks lines of code as inactive.
         */
        void markInactiveCode(TokenData td)
        {
            var fncdata = NCMCSParser.structdata.functionGetData();
            if (fncdata == null || fncdata.counterList == null)
            {
                return;
            }

            for (int current = td.StartLocation.Line; current <= td.EndLocation.Line; current++)
            {
                addTargetLine(ref fncdata.counterList, current);
                fncdata.counterList[current].cntEloc = 0;
                fncdata.counterList[current].isInactiveCode = true;

            }
        }


        /**
         * カウント用リストの作成
         * Initializes the list
         * that holds metric data.
         */
        bool preLineCounter()
        {
            NCMStructData.NCMFuncData fncdata = null;
            fncdata = NCMCSParser.structdata.functionGetData();


            if (fncdata != null && fncdata.counterList == null)
            {
                fncdata.counterList = new Dictionary<int, NCMStructData.NCMFuncData.NCMLineData>();
                return true;
            }
            return false;
        }

        /**
          * メソッド系ビジター毎にラインカウントする
          * Counts the metric value after each
          * visitor method.
          */
        void mainLineCounter(NCMStructData.NCMFuncData fncdata)
        {
            if (fncdata == null || fncdata.counterList == null)
            {
                return;
            }

            foreach (var item in fncdata.counterList)
            {
                if (item.Value.cntEloc != 0)
                {
                    fncdata.cntEloc++;
                }
                else
                {
                    if (item.Value.cntCloc != 0)
                    {
                        fncdata.cntCloc++;
                    }

                    if (item.Value.isInactiveCode)
                    {
                        fncdata.cntSkip++;
                    }

                }
            }

            debugPrintLineDetail(fncdata);


            fncdata.counterList.Clear();
            fncdata.counterList = null;

        }


        #region FUNC用クラス名リスト

        /**
         * 関数名リストから関数名を取得.
         * This function is used to get the name of the function with
         * its parent(s) from the list.
         */
        public String getScopeNames()
        {
            String txt = "";
            foreach (String fnm in scopenames)
                txt += fnm + ".";

            if (txt.Length > 1)
                txt = txt.Remove(txt.Length - 1);

            return txt;
        }

        /**
         * 関数名リストの最後尾に関数名を追加.
         * This function is used to add the
         * name of the function with all
         * its parent(s).
         */
        private void addScopeList(String name)
        {
            scopenames.Add(name);
        }

        /**
         * 関数名リストの最後尾を削除.
         * This function is used to
         * remove the name present
         * at the top.
         */
        private void removeScopeList()
        {
            if (scopenames.Count > 0)
            {
                scopenames.RemoveAt(scopenames.Count - 1);
            }
        }

        /**
         * 逐語的識別子(@プレフィックス)が指定されている場合、snmに"@"を付加する。
         * This function is used add '@' to the name of the string if verbitam
         * name if present.
         */
        string GetVerbatimName(string snm, SyntaxToken identifier)
        {
            return (identifier == null) ? snm : (identifier.IsVerbatimIdentifier()) ? "@" + snm : snm;
        }


        #endregion

        #region クラス系ノード（Class,Struct,Delegate）

        /**
         * クラスノード。名前を関数名リストに登録する.
         * This function is used to visit the Class and add its
         * name in the list.
         */
        public override void VisitClassDeclaration(ClassDeclarationSyntax classDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000051_01); //* MSG_LOG_ID_CS_00000051
            if (!flgOutsideFunction
                && classDeclaration.Identifier != null)
            {
                String sname = classDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, classDeclaration.Identifier);


                addScopeList(sname);
            }

            base.VisitClassDeclaration(classDeclaration);

            if (!flgOutsideFunction)
            {
                removeScopeList();
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000052_01); //* MSG_LOG_ID_CS_00000052
        }

        /**
         * クラスノード。名前を関数名リストに登録する.
         * This function is used to visit the Struct
         * construct and add its name in the list.
         */
        public override void VisitStructDeclaration(StructDeclarationSyntax structDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000051_01); //* MSG_LOG_ID_CS_00000051
            if (!flgOutsideFunction
                && structDeclaration.Identifier != null)
            {
                String sname = structDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, structDeclaration.Identifier);


                addScopeList(sname);
            }

            base.VisitStructDeclaration(structDeclaration);

            if (!flgOutsideFunction)
            {
                removeScopeList();
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000052_01); //* MSG_LOG_ID_CS_00000052
        }

        /**
         * クラスノード。名前を関数名リストに登録する.
         * This function is used to visit the Interface
         * construct and add its name in the list.
         */
        public override void VisitInterfaceDeclaration(InterfaceDeclarationSyntax interfaceDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000051_01); //* MSG_LOG_ID_CS_00000051
            if (!flgOutsideFunction
                && interfaceDeclaration.Identifier != null)
            {
                String sname = interfaceDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, interfaceDeclaration.Identifier);


                addScopeList(sname);
            }

            base.VisitInterfaceDeclaration(interfaceDeclaration);

            if (!flgOutsideFunction)
            {
                removeScopeList();
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000052_01); //* MSG_LOG_ID_CS_00000052
        }

        /**
        * イベントアクセサーノード。名前を関数名リストに登録する.
        * This function is used to visit Event construct and
        * add its name in the list.
        */
        public override void VisitEventDeclaration(EventDeclarationSyntax eventDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000053_01); //* MSG_LOG_ID_CS_00000053

            if (!flgOutsideFunction
                && eventDeclaration.Identifier != null)
            {
                String sname = eventDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, eventDeclaration.Identifier);


                if (eventDeclaration.ExplicitInterfaceSpecifier != null)
                {

                    sname = eventDeclaration.ExplicitInterfaceSpecifier.Name + "." + sname;
                }
                addScopeList(sname);


                commentLineList = new List<int>();

            }
            base.VisitEventDeclaration(eventDeclaration);

            if (!flgOutsideFunction)
            {
                removeScopeList();


                commentLineList.Clear();
                commentLineList = null;

            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000054_01); //* MSG_LOG_ID_CS_00000054

        }

        /**
        * インデクサノード。名前(this)を関数名リストに登録する.
        * This function is used to visit Indexer construct
        * and add "this" in the name list.
        */
        public override void VisitIndexerDeclaration(IndexerDeclarationSyntax indexerDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000055_01); //* MSG_LOG_ID_CS_00000055

            if (!flgOutsideFunction)
            {

                String sname = "this";
                if (indexerDeclaration.ThisKeyword != null)
                {
                    sname = indexerDeclaration.ThisKeyword.Text;


                    sname = GetVerbatimName(sname, indexerDeclaration.ThisKeyword);

                }
                if (indexerDeclaration.ExplicitInterfaceSpecifier != null)
                {
                    sname = indexerDeclaration.ExplicitInterfaceSpecifier.Name + "." + sname;
                }
                addScopeList(sname);


                commentLineList = new List<int>();

            }


            base.VisitIndexerDeclaration(indexerDeclaration);


            if (!flgOutsideFunction)
            {
                removeScopeList();


                commentLineList.Clear();
                commentLineList = null;

            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000056_01); //* MSG_LOG_ID_CS_00000056
        }

        /**
         * プロパティノード。名前を関数名リストに登録する.
         * This function is used to visit Property
         * construct and add its name in the list.
         */
        public override void VisitPropertyDeclaration(PropertyDeclarationSyntax propertyDeclaration)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000057_01); //* MSG_LOG_ID_CS_00000057

            if (!flgOutsideFunction
                && propertyDeclaration.Identifier != null)
            {
                String sname = propertyDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, propertyDeclaration.Identifier);



                if (propertyDeclaration.ExplicitInterfaceSpecifier != null)
                {
                    sname = propertyDeclaration.ExplicitInterfaceSpecifier.Name + "." + sname;
                }
                addScopeList(sname);
            }

            base.VisitPropertyDeclaration(propertyDeclaration);

            if (!flgOutsideFunction)
            {
                removeScopeList();
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000058_01); //* MSG_LOG_ID_CS_00000058

        }

        #endregion

        #region メソッド系ノード

        /**
         * パラメータの型名を取得。メソッド系処理で利用。
         * Returns the names of the parameters
         * enclosed withing parenthesis.
         */
        string GetParenthesis(IEnumerable<SyntaxNode> list)
        {
            string res = "(";
            bool isFirst = true;
            if (list.Any())
            {
                foreach (ParameterSyntax node in list.OfType<ParameterSyntax>())
                {
                    if (isFirst)
                    {
                        isFirst = false;
                    }
                    else
                    {
                        res += ",";
                    }

                    if (node.Type == null)
                    {
                        res += " " + "__arglist" + " ";
                    }
                    else
                    {


                        res += " " + node.Type.ToString() + " ";
                    }
                }
            }
            else
            {

                res += " ";
            }
            return res + ")";
        }

        /**
         * コンストラクタノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）を実行。
         * This function used to visit Constructor method in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitConstructorDeclaration(ConstructorDeclarationSyntax constructorDeclaration)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000059_01); //* MSG_LOG_ID_CS_00000059

            if (!flgOutsideFunction
                && constructorDeclaration.Body != null
                && !constructorDeclaration.Body.IsMissing
                && constructorDeclaration.Identifier != null)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = constructorDeclaration.GetLocation().GetLineSpan().StartLinePosition.Line;
                String sname = constructorDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, constructorDeclaration.Identifier);


                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(constructorDeclaration.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(constructorDeclaration.ParameterList.Parameters);

                fncdata.FuncStartToken = constructorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.cntFunctionBase++;


                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000060_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000060_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000060_03); //* MSG_LOG_ID_CS_00000060


                preLineCounter();
            }

            base.VisitConstructorDeclaration(constructorDeclaration);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (constructorDeclaration.Body != null) ? constructorDeclaration.Body.GetLocation().GetLineSpan().StartLinePosition : constructorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = (constructorDeclaration.Body != null) ? constructorDeclaration.Body.GetLocation().GetLineSpan().EndLinePosition : constructorDeclaration.GetLocation().GetLineSpan().EndLinePosition;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000061_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000061_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000061
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000062
            }


        NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000063_01); //* MSG_LOG_ID_CS_00000063

        }

        /**
         * デストラクターノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * ※デストラクターはパラメーター、アクセシビリティ修飾子の指定は行えない。
         * This function used to visit Destructor methond in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitDestructorDeclaration(DestructorDeclarationSyntax destructorDeclaration)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000064_01); //* MSG_LOG_ID_CS_00000064

            if (!flgOutsideFunction
                && destructorDeclaration.Body != null
                && !destructorDeclaration.Body.IsMissing
                && destructorDeclaration.Identifier != null)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = destructorDeclaration.GetLocation().GetLineSpan().StartLinePosition.Line;

                TypeDeclarationSyntax type = destructorDeclaration.Parent as TypeDeclarationSyntax;


                String sname = string.Empty;
                if (type != null && type.Identifier != null)
                {
                    sname = GetVerbatimName(type.Identifier.ValueText, type.Identifier);
                }
                else
                {
                    sname = GetVerbatimName(destructorDeclaration.Identifier.ValueText, destructorDeclaration.Identifier);
                }


                if (destructorDeclaration.TildeToken.Kind() == SyntaxKind.TildeToken)
                {
                    sname = "~" + sname;
                }
                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(destructorDeclaration.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = "( )";

                fncdata.FuncStartToken = destructorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.cntFunctionBase++;




                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000065_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000065_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000065_03); //* MSG_LOG_ID_CS_00000065


                preLineCounter();
            }

            base.VisitDestructorDeclaration(destructorDeclaration);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (destructorDeclaration.Body != null) ? destructorDeclaration.Body.GetLocation().GetLineSpan().StartLinePosition : destructorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = (destructorDeclaration.Body != null) ? destructorDeclaration.Body.GetLocation().GetLineSpan().EndLinePosition : destructorDeclaration.GetLocation().GetLineSpan().EndLinePosition;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000066_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000066_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000066
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000067

            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000068_01); //* MSG_LOG_ID_CS_00000068

        }

        /**
         * メソッドノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * This function used to visit all methond in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitMethodDeclaration(MethodDeclarationSyntax methodDeclaration)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000069_01); //* MSG_LOG_ID_CS_00000069

            if (!flgOutsideFunction
                /** For expression bodied method
                 *  declarations, the 'Body' of the 
                 *  method is null and hence 
                 *  'ExpressionBody' field should be 
                 *  used instead.
                 */
                && ((methodDeclaration.Body != null
                && !methodDeclaration.Body.IsMissing)
                || methodDeclaration.ExpressionBody != null)
                && methodDeclaration.Identifier != null)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = methodDeclaration.GetLocation().GetLineSpan().StartLinePosition.Line;
                String sname = methodDeclaration.Identifier.ValueText;


                sname = GetVerbatimName(sname, methodDeclaration.Identifier);

                if (methodDeclaration.ExplicitInterfaceSpecifier != null)
                {
                    sname = methodDeclaration.ExplicitInterfaceSpecifier.Name + "." + sname;
                }

                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(methodDeclaration.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(methodDeclaration.ParameterList.Parameters);

                fncdata.FuncStartToken = methodDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.cntFunctionBase++;




                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000070_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000070_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000070_03); //* MSG_LOG_ID_CS_00000070


                preLineCounter();
            }

            base.VisitMethodDeclaration(methodDeclaration);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = methodDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = methodDeclaration.GetLocation().GetLineSpan().EndLinePosition;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000071_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000071_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000071
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000072


            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000073_01); //* MSG_LOG_ID_CS_00000073

        }

        /**
         * アクセサノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * This function used to visit Accessor in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitAccessorDeclaration(AccessorDeclarationSyntax accessor)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000074_01); //* MSG_LOG_ID_CS_00000074

#if !DEBUG_AUTO_PROPRETY
            if (!flgOutsideFunction
                && accessor.Keyword != null)
#else

            if (!flgOutsideFunction
                && !accessor.Body.IsNull
                && accessor.Body != null
                && accessor.Keyword != null)
#endif
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = accessor.GetLocation().GetLineSpan().StartLinePosition.Line;
                String sname = string.Empty;
                if (accessor.Keyword.Kind() == SyntaxKind.GetKeyword)
                {
                    sname = "get";
                }
                else if (accessor.Keyword.Kind() == SyntaxKind.SetKeyword)
                {
                    sname = "set";
                }
                else if (accessor.Keyword.Kind() == SyntaxKind.AddKeyword)
                {
                    sname = "add";
                }
                else if (accessor.Keyword.Kind() == SyntaxKind.RemoveKeyword)
                {
                    sname = "remove";
                }
                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(accessor.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = string.Empty;

#if DEBUG_ACCESSOR_ATTRIBUTES
                int attributeCount = accessor.Attributes.Count;
                fncdata.FuncStartToken =accessor.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.FuncStartToken.Line += attributeCount;
                fncdata.cntEloc = -1 * attributeCount;
#else
                fncdata.FuncStartToken = accessor.GetLocation().GetLineSpan().StartLinePosition;
#endif

                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000075_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000075_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000075_03); //* MSG_LOG_ID_CS_00000075


                preLineCounter();

            }

            base.VisitAccessorDeclaration(accessor);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (accessor.Body != null && !accessor.Body.IsMissing) ? accessor.Body.GetLocation().GetLineSpan().StartLinePosition :
                                         accessor.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = (accessor.Body != null && !accessor.Body.IsMissing) ? accessor.Body.GetLocation().GetLineSpan().EndLinePosition :
                                       accessor.GetLocation().GetLineSpan().EndLinePosition;
                if (commentLineList != null)
                {
                    /* Mark the lines contained in commentLineList
                     * as comments and increment CLOC.
                     */
                    foreach (int currentLine in commentLineList)
                    {
                        if (currentLine < fncdata.BodyStartToken.Line ||
                           currentLine > fncdata.BodyEndToken.Line)
                        {
                            continue;
                        }

                        if (!fncdata.counterList.ContainsKey(currentLine))
                        {
                            addTargetLine(ref fncdata.counterList, currentLine);
                            fncdata.counterList[currentLine].cntCloc++;
                        }
                    }
                }
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000076_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000076_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000076
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000077

            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000078_01); //* MSG_LOG_ID_CS_00000078
        }

        /**
         * オペレータノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * This function used to visit Operator in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitConversionOperatorDeclaration(ConversionOperatorDeclarationSyntax conversionOperatorDeclaration)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000079_01); //* MSG_LOG_ID_CS_00000079

            if (!flgOutsideFunction
                /** For expression bodied implicit/explicit
                 *  operator declarations, the 'Body' of the
                 *  method is null and hence 
                 *  'ExpressionBody' field should be 
                 *  used instead.
                 */
                && ((conversionOperatorDeclaration.Body != null
                && conversionOperatorDeclaration.Body.Statements != null)
                || conversionOperatorDeclaration.ExpressionBody != null))
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = conversionOperatorDeclaration.GetLocation().GetLineSpan().StartLinePosition.Line;
                String sname = "operator";

                sname += conversionOperatorDeclaration.ImplicitOrExplicitKeyword.Kind() == SyntaxKind.ImplicitKeyword
                    || conversionOperatorDeclaration.ImplicitOrExplicitKeyword.Kind() == SyntaxKind.ExplicitKeyword ? " "
                    + conversionOperatorDeclaration.Type.GetText() : " " + conversionOperatorDeclaration.OperatorKeyword.Text;
                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(conversionOperatorDeclaration.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(conversionOperatorDeclaration.ParameterList.Parameters);

                fncdata.FuncStartToken = conversionOperatorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.cntFunctionBase++;

                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000080_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000080_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000080_03); //* MSG_LOG_ID_CS_00000080


                preLineCounter();
            }

            base.VisitConversionOperatorDeclaration(conversionOperatorDeclaration);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (conversionOperatorDeclaration.Body != null) ? conversionOperatorDeclaration.Body.GetLocation().GetLineSpan().StartLinePosition : conversionOperatorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = (conversionOperatorDeclaration.Body != null) ? conversionOperatorDeclaration.Body.GetLocation().GetLineSpan().EndLinePosition : conversionOperatorDeclaration.GetLocation().GetLineSpan().EndLinePosition;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000081_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000081_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000081
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000082


            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000083_01); //* MSG_LOG_ID_CS_00000083


        }



        /**
         * オペレータノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * This function used to visit Operator in the source file and
         * set all the data related to the method, like: position, BODY start position,
         * end position etc.
         */
        public override void VisitOperatorDeclaration(OperatorDeclarationSyntax operatorDeclaration)
        {
            int iline = 0;
            bool flgStartFunc = false;
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000079_01); //* MSG_LOG_ID_CS_00000079

            if (!flgOutsideFunction
                /** For expression bodied operator
                 *  declarations, the 'Body' of the 
                 *  operator is null and hence 
                 *  'ExpressionBody' field should be 
                 *  used instead.
                 */
                && ((operatorDeclaration.Body != null
                && operatorDeclaration.Body.Statements != null)
                || operatorDeclaration.ExpressionBody != null))
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = operatorDeclaration.GetLocation().GetLineSpan().StartLinePosition.Line;
                String sname = "operator";

                sname += operatorDeclaration.OperatorToken.Kind() == SyntaxKind.ImplicitKeyword
                    || operatorDeclaration.OperatorToken.Kind() == SyntaxKind.ExplicitKeyword ? " " + operatorDeclaration.ReturnType.ToString() : " " + operatorDeclaration.OperatorToken.Text;
                addScopeList(sname);

                fncdata = NCMCSParser.structdata.functionNewData(operatorDeclaration.GetLocation().GetLineSpan().StartLinePosition);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(operatorDeclaration.ParameterList.Parameters);

                fncdata.FuncStartToken = operatorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.cntFunctionBase++;

                NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000080_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000080_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000080_03); //* MSG_LOG_ID_CS_00000080


                preLineCounter();
            }

            base.VisitOperatorDeclaration(operatorDeclaration);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (operatorDeclaration.Body != null) ? operatorDeclaration.Body.GetLocation().GetLineSpan().StartLinePosition : operatorDeclaration.GetLocation().GetLineSpan().StartLinePosition;
                fncdata.BodyEndToken = (operatorDeclaration.Body != null) ? operatorDeclaration.Body.GetLocation().GetLineSpan().EndLinePosition : operatorDeclaration.GetLocation().GetLineSpan().EndLinePosition;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata);

                NCMCSParser.structdata.dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000081_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000081_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000081
                NCMCSParser.structdata.dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000082
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000083_01); //* MSG_LOG_ID_CS_00000083

        }

        #endregion

        #region メトリクスカウント用 Override

        /**
          * ブロック数を測定。
          * This function measures the
          * number of code blocks.
          */
        public override void VisitBlock(BlockSyntax blockStatement)
        {
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000084_01); //* MSG_LOG_ID_CS_00000084

            if (flgOutsideFunction)
            {
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {

                    if (!flgFunctionBlock)
                    {

                        fncdata.cntBranchs++;
                        fncdata.cntCyclo++;
                        fncdata.cntCycloWC++;
                        flgFunctionBlock = true;
                    }

                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000085_01 + blockStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_02 + blockStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_03 + blockStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_04); //* MSG_LOG_ID_CS_00000085
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000086_01); //* MSG_LOG_ID_CS_00000086
                }
            }

            base.VisitBlock(blockStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000087_01); //* MSG_LOG_ID_CS_00000087

        }

        /**
          * ForEachノード。CYCLO,BRANCHSのカウント。
          * Visitor method for for-each constructs.
          */
        public override void VisitForEachStatement(ForEachStatementSyntax foreachStatement)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000088_01); //* MSG_LOG_ID_CS_00000088
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsFor++;
                    fncdata.cntBranchs++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000089_01 + foreachStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_02 + foreachStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_03 + foreachStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_04); //* MSG_LOG_ID_CS_00000089


                    addTargetLine(ref fncdata.counterList, foreachStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[foreachStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntLoopsFor++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000090_01); //* MSG_LOG_ID_CS_00000090
                }
            }

            base.VisitForEachStatement(foreachStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000091_01); //* MSG_LOG_ID_CS_00000091

        }

        /**
         * Forステートメントノード。CYCLO,BRANCHSのカウント。
         * Visitor method for For loops.
         */
        public override void VisitForStatement(ForStatementSyntax forStatement)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000092_01); //* MSG_LOG_ID_CS_00000092
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsFor++;
                    fncdata.cntBranchs++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000093_01 + forStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_02 + forStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_03 + forStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_04); //* MSG_LOG_ID_CS_00000093


                    flgBranchCondition = true;
                    if ( forStatement.Condition != null )
                        forStatement.Condition.Accept(this);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, forStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[forStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntLoopsFor++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000094_01); //* MSG_LOG_ID_CS_00000094
                }
            }

            base.VisitForStatement(forStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000095_01); //* MSG_LOG_ID_CS_00000095

        }

        /**
         * Switchステートメントノード。CYCLOのカウント。
         * Visitor for switch statement.
         */
        public override void VisitSwitchStatement(SwitchStatementSyntax switchStatement)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000096_01); //* MSG_LOG_ID_CS_00000096
            NCMStructData.NCMFuncData fncdata = null;
            if (flgOutsideFunction && !flgBranchCondition)
            {

                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {

                    fncdata.cntSelectionSwitch++;
                    fncdata.cntCyclo++;

                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000097_01 + switchStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_02 + switchStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_03 + switchStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_04); //* MSG_LOG_ID_CS_00000097


                    flgBranchCondition = true;
                    switchStatement.Expression.Accept(this);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, switchStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[switchStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntSelectionSwitch++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000098_01); //* MSG_LOG_ID_CS_00000098
                }
            }

            base.VisitSwitchStatement(switchStatement);
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000099_01); //* MSG_LOG_ID_CS_00000099


        }

        /**
         * Caseラベルノード。BRANCHSのカウント。
         * Visitor method for cases inside switch
         * statements/blocks.
         */
        public override void VisitCaseSwitchLabel(CaseSwitchLabelSyntax caseLabel)
        {

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000100_01); //* MSG_LOG_ID_CS_00000100
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {

                    if (caseLabel.Value == null)
                    {

                        fncdata.cntSelectionDefault++;
                    }
                    else
                    {

                        fncdata.cntSelectionCase++;
                        fncdata.cntBranchs++;
                        fncdata.cntCycloWC++;


                        flgBranchCondition = true;
                        if ( caseLabel.Value != null )
                            caseLabel.Value.Accept(this);
                        flgBranchCondition = false;
                    }
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000101_01 + caseLabel.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_02 + caseLabel.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_03 + caseLabel.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_04); //* MSG_LOG_ID_CS_00000101


                    addTargetLine(ref fncdata.counterList, caseLabel.GetLocation().GetLineSpan().StartLinePosition.Line);
                    if (caseLabel.Value == null)
                    {
                        fncdata.counterList[caseLabel.GetLocation().GetLineSpan().StartLinePosition.Line].cntSelectionDefault++;
                    }
                    else
                    {
                        fncdata.counterList[caseLabel.GetLocation().GetLineSpan().StartLinePosition.Line].cntSelectionCase++;
                    }
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000102_01); //* MSG_LOG_ID_CS_00000102
                }
            }


            base.VisitCaseSwitchLabel(caseLabel);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000103_01); //* MSG_LOG_ID_CS_00000103

        }

        /**
         * Doループステートメントノード。CYCLO,BRANCHSのカウント。
         * Visitor method for do-while loops/constructs.
         */
        public override void VisitDoStatement(DoStatementSyntax doWhileStatement)
        {

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000104_01); //* MSG_LOG_ID_CS_00000104
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsDoWhile++;
                    fncdata.cntBranchs++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000105_01 + doWhileStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_02 + doWhileStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_03 + doWhileStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_04); //* MSG_LOG_ID_CS_00000105


                    flgBranchCondition = true;
                    if ( doWhileStatement.Condition != null )
                        doWhileStatement.Condition.Accept(this);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, doWhileStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[doWhileStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntLoopsDoWhile++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000106_01); //* MSG_LOG_ID_CS_00000106
                }
            }

            base.VisitDoStatement(doWhileStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000107_01); //* MSG_LOG_ID_CS_00000107

        }

        /**
         * Whileループステートメントノード。CYCLO,BRANCHSのカウント。
         * Visitor method for while loop/constructs.
         */
        public override void VisitWhileStatement(WhileStatementSyntax whileStatement)
        {

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000108_01); //* MSG_LOG_ID_CS_00000108
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsDoWhile++;
                    fncdata.cntBranchs++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000109_01 + whileStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_02 + whileStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_03 + whileStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_04); //* MSG_LOG_ID_CS_00000109


                    flgBranchCondition = true;
                    if ( whileStatement.Condition != null )
                        whileStatement.Condition.Accept(this);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, whileStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[whileStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntLoopsDoWhile++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000110_01); //* MSG_LOG_ID_CS_00000110
                }
            }

            base.VisitWhileStatement(whileStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000111_01); //* MSG_LOG_ID_CS_00000111

        }

        /**
         * ifステートメントノード。CYCLO,BRANCHSのカウント。
         * Visitor method for If statements.
         */
        public override void VisitIfStatement(IfStatementSyntax ifElseStatement)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000112_01); //* MSG_LOG_ID_CS_00000112
            if (flgOutsideFunction && !flgBranchCondition)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntConditionalIf++;
                    fncdata.cntBranchs++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000113_01 + ifElseStatement.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_02 + ifElseStatement.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_03 + ifElseStatement.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_04); //* MSG_LOG_ID_CS_00000113


                    flgBranchCondition = true;
                    if ( ifElseStatement.Condition != null )
                        ifElseStatement.Condition.Accept(this);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, ifElseStatement.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[ifElseStatement.GetLocation().GetLineSpan().StartLinePosition.Line].cntConditionalIf++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000114_01); //* MSG_LOG_ID_CS_00000114
                }
            }

            base.VisitIfStatement(ifElseStatement);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000115_01); //* MSG_LOG_ID_CS_00000115

        }

        /**
         * Function to print debug lines for binary operator expression
         */

        public void PrintDebugLines(BinaryExpressionSyntax binaryOperatorExpression)
        {
            NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000117_01 + binaryOperatorExpression.Kind().ToString()
            + Resources.MSG_LOG_ID_CS_00000117_02 + binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
            + Resources.MSG_LOG_ID_CS_00000117_03 + binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
            + Resources.MSG_LOG_ID_CS_00000117_04); //* MSG_LOG_ID_CS_00000117    

        }

        /**
          * 演算式ノード。BRANCHS, CNDOPS の（&& , || や Null Coalesce）カウント。
          * Visitor method for Conditional 'AND', 'OR' and Null Coalesce operator.
          */
        public override void VisitBinaryExpression(BinaryExpressionSyntax binaryOperatorExpression)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000116_01); //* MSG_LOG_ID_CS_00000116

            
            /** flgOutsideFunction is true for all - AND, OR and Null Coalesce */
            if (flgOutsideFunction)
            {
                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    switch (binaryOperatorExpression.Kind())
                    {

                        /** For Null coalesce kind of binary expression is CoalesceExpression */
                        case SyntaxKind.CoalesceExpression:
                            /**
                                For Null Coalesce operator
                                      1. flgBranchCondition is not true 
                                      2. cntCndOps increases
                                      3. cntInlinedIf increases
                                above 3 conditions are taken from ternary operator case
                                because null coalesce semantically behaves as ternary operator
                            */
                            if (!flgBranchCondition)
                            {
                                fncdata.cntInlinedIf++;
                                fncdata.cntCndOps++;
                                addTargetLine(ref fncdata.counterList, binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Line);
                                fncdata.counterList[binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Line].cntCndOps++;
                                /** to print debug lines */
                                PrintDebugLines(binaryOperatorExpression);
                            }
                            
                            break;
                            
                        case SyntaxKind.LogicalAndExpression:
                        case SyntaxKind.LogicalOrExpression:
                                
                            if (flgBranchCondition)
                            {
                                fncdata.cntLogicalAnd++;
                                fncdata.cntBranchs++;
                                addTargetLine(ref fncdata.counterList, binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Line);
                                fncdata.counterList[binaryOperatorExpression.GetLocation().GetLineSpan().StartLinePosition.Line].cntLogicalAnd++;
                                /** to print debug lines */
                                PrintDebugLines(binaryOperatorExpression);

                            }
                            
                            break;

                        default:

                            break;
                    } 
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000118_01); //* MSG_LOG_ID_CS_00000118
                }

            }

            base.VisitBinaryExpression(binaryOperatorExpression);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000119_01); //* MSG_LOG_ID_CS_00000119
        }


        /**
         * Null条件式ノード。 CNDOPSのカウント。
         * Visitor method for Null Conditional expression.
         */


        public override void VisitConditionalAccessExpression(ConditionalAccessExpressionSyntax conditionalAccessExpression)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000120_01); //* MSG_LOG_ID_CS_00000120
            /** These conditions are same as for ternary operator case */
            if (flgOutsideFunction && !flgBranchCondition)
            {

                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    /** Increased these 2 metrics as these 2 are only getting incremented in ternary operator case */
                    fncdata.cntInlinedIf++;
                    fncdata.cntCndOps++;


                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000121_01
                    + conditionalAccessExpression.Expression.ToString() + Resources.MSG_LOG_ID_CS_00000121_02
                    + conditionalAccessExpression.WhenNotNull.ToString() + Resources.MSG_LOG_ID_CS_00000121_03
                    + conditionalAccessExpression.WhenNotNull.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_04 + conditionalAccessExpression.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_05 + conditionalAccessExpression.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_06); //* MSG_LOG_ID_CS_00000121

                    addTargetLine(ref fncdata.counterList, conditionalAccessExpression.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[conditionalAccessExpression.GetLocation().GetLineSpan().StartLinePosition.Line].cntCndOps++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000122_01); //* MSG_LOG_ID_CS_00000122
                }
            }

            base.VisitConditionalAccessExpression(conditionalAccessExpression);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000123_01); //* MSG_LOG_ID_CS_00000123
        }
        




        /**
         * 三項演算式ノード。CNDOPSのカウント。
         * Visitor method for ternary expression.
         */
        public override void VisitConditionalExpression(ConditionalExpressionSyntax conditionalExpression)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000120_01); //* MSG_LOG_ID_CS_00000120

            if (flgOutsideFunction && !flgBranchCondition)
            {

                NCMStructData.NCMFuncData fncdata = null;
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntInlinedIf++;
                    fncdata.cntCndOps++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000121_01
                    + conditionalExpression.Condition.ToString() + Resources.MSG_LOG_ID_CS_00000121_02
                    + conditionalExpression.WhenTrue.ToString() + Resources.MSG_LOG_ID_CS_00000121_03
                    + conditionalExpression.WhenFalse.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_04 + conditionalExpression.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_05 + conditionalExpression.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_06); //* MSG_LOG_ID_CS_00000121

                    addTargetLine(ref fncdata.counterList, conditionalExpression.GetLocation().GetLineSpan().StartLinePosition.Line);
                    fncdata.counterList[conditionalExpression.GetLocation().GetLineSpan().StartLinePosition.Line].cntCndOps++;
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000122_01); //* MSG_LOG_ID_CS_00000122
                }
            }

            base.VisitConditionalExpression(conditionalExpression);

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000123_01); //* MSG_LOG_ID_CS_00000123
        }

        public override void VisitLiteralExpression(LiteralExpressionSyntax nullReferenceExpression)
        {
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000126_01); //* MSG_LOG_ID_CS_00000126
            base.VisitLiteralExpression(nullReferenceExpression);

            if (flgOutsideFunction)
            {
                TokenData td = new TokenData("", nullReferenceExpression.GetType().ToString(),
                                                nullReferenceExpression.Kind(),
                                                nullReferenceExpression.GetLocation().GetLineSpan().StartLinePosition,
                                                nullReferenceExpression.GetLocation().GetLineSpan().EndLinePosition);

                CountToken(td);
            }

            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000127_01); //* MSG_LOG_ID_CS_00000127
        }

        /** Visitor for the expression body of an expression bodied member. */
        public override void VisitArrowExpressionClause(ArrowExpressionClauseSyntax expressionBodiedRepresentation)
        {
            NCMStructData.NCMFuncData fncdata = null;
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000128_01); //* MSG_LOG_ID_CS_000000128
            if (flgOutsideFunction)
            {
                fncdata = NCMCSParser.structdata.functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000129_01 + expressionBodiedRepresentation.Kind().ToString()
                    + Resources.MSG_LOG_ID_CS_00000129_02 + expressionBodiedRepresentation.GetLocation().GetLineSpan().StartLinePosition.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000129_03 + expressionBodiedRepresentation.GetLocation().GetLineSpan().StartLinePosition.Character.ToString()
                    + Resources.MSG_LOG_ID_CS_00000129_04); //* MSG_LOG_ID_CS_00000129
                }
                else
                {
                    NCMCSParser.structdata.dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000130_01); //* MSG_LOG_ID_CS_00000130
                }
            }
            base.VisitArrowExpressionClause(expressionBodiedRepresentation);
            NCMCSParser.structdata.dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000131_01); //* MSG_LOG_ID_CS_00000131
        }

        #endregion

        #region Statement系ノード
        public override void VisitEmptyStatement(EmptyStatementSyntax emptyStatement)
        {
            TokenData td = new TokenData("", emptyStatement.GetType().ToString(), emptyStatement.Kind(),
                                        emptyStatement.GetLocation().GetLineSpan().StartLinePosition,
                                        emptyStatement.GetLocation().GetLineSpan().EndLinePosition);


            CountToken(td);

            base.VisitEmptyStatement(emptyStatement);
        }
        #endregion

        #region Token系ノード

        /**
         * Default visitor for all the token nodes.
         * Calculates metric values for specific
         * nodes and then visits its children.
         */
        public override void VisitToken(SyntaxToken token)
        {
            /* Presence of curly braces in Interpolated Strings
             * should not trigger a change in the NESTS metric
             * therefore processing of such braces is avoided.
             */
            if(token.RawKind == (int) SyntaxKind.OpenBraceToken && token.Parent.RawKind == (int)SyntaxKind.Interpolation)
            {
                return;
            }
            TokenData td = new TokenData("", token.GetType().ToString(),
                               token.Kind(), token.GetLocation().GetLineSpan().StartLinePosition, token.GetLocation().GetLineSpan().EndLinePosition);
            CountToken(td);
            base.VisitToken(token);
        }

        public override void VisitPredefinedType(PredefinedTypeSyntax primitiveType)
        {
            TokenData td = new TokenData(primitiveType.GetText().ToString(), primitiveType.GetType().ToString(),
                               primitiveType.Kind(),
                               primitiveType.GetLocation().GetLineSpan().StartLinePosition,
                               primitiveType.GetLocation().GetLineSpan().EndLinePosition);
            td.RoleName = primitiveType.Kind().ToString();

            CountToken(td);

            base.VisitPredefinedType(primitiveType);
        }


        /**
         * Visitor method for Comments.
         */
        public void VisitComment(SyntaxTrivia comment)
        {
            TokenData td = new TokenData(comment.ToString(), comment.GetType().ToString(),
                               comment.Kind(), comment.GetLocation().GetLineSpan().StartLinePosition,
                               comment.GetLocation().GetLineSpan().EndLinePosition);
            td.ItemType = comment.Kind().ToString();
            FileLinePositionSpan location = comment.GetLocation().GetLineSpan();



            if (commentLineList != null)
            {
                for (int i = location.StartLinePosition.Line; i <= location.EndLinePosition.Line; i++)
                {
                    commentLineList.Add(i);
                }
            }

            CountToken(td);
        }

        /**
         * Visitor method for  '#if'
         * directive.
         */
        public override void VisitIfDirectiveTrivia(IfDirectiveTriviaSyntax ifDirective)
        {
            TokenData td = new TokenData(ifDirective.ToString(), ifDirective.GetType().ToString(),
                               ifDirective.Kind(), ifDirective.GetLocation().GetLineSpan().StartLinePosition,
                               ifDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = ifDirective.GetType().ToString();
            td.PpTake = ifDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#elif'
         * directive.
         */
        public override void VisitElifDirectiveTrivia(ElifDirectiveTriviaSyntax elifDirective)
        {
            TokenData td = new TokenData(elifDirective.ToString(), elifDirective.GetType().ToString(),
                               elifDirective.Kind(), elifDirective.GetLocation().GetLineSpan().StartLinePosition,
                               elifDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = elifDirective.GetType().ToString();
            td.PpTake = elifDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#else'
         * directive.
         */
        public override void VisitElseDirectiveTrivia(ElseDirectiveTriviaSyntax elseDirective)
        {
            TokenData td = new TokenData(elseDirective.ToString(), elseDirective.GetType().ToString(),
                               elseDirective.Kind(), elseDirective.GetLocation().GetLineSpan().StartLinePosition,
                               elseDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = elseDirective.GetType().ToString();
            td.PpTake = elseDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#endif'
         * directive.
         */
        public override void VisitEndIfDirectiveTrivia(EndIfDirectiveTriviaSyntax endifDirective)
        {
            TokenData td = new TokenData(endifDirective.ToString(), endifDirective.GetType().ToString(),
                               endifDirective.Kind(), endifDirective.GetLocation().GetLineSpan().StartLinePosition,
                               endifDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = endifDirective.GetType().ToString();
            td.PpTake = endifDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#region'
         * directive.
         */
        public override void VisitRegionDirectiveTrivia(RegionDirectiveTriviaSyntax regionDirective)
        {
            TokenData td = new TokenData(regionDirective.ToString(), regionDirective.GetType().ToString(),
                               regionDirective.Kind(), regionDirective.GetLocation().GetLineSpan().StartLinePosition,
                               regionDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = regionDirective.GetType().ToString();
            td.PpTake = regionDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#endregion'
         * directive.
         */
        public override void VisitEndRegionDirectiveTrivia(EndRegionDirectiveTriviaSyntax endRegionDirective)
        {
            TokenData td = new TokenData(endRegionDirective.ToString(), endRegionDirective.GetType().ToString(),
                               endRegionDirective.Kind(), endRegionDirective.GetLocation().GetLineSpan().StartLinePosition,
                               endRegionDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = endRegionDirective.GetType().ToString();
            td.PpTake = endRegionDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#define'
         * directive.
         */
        public override void VisitDefineDirectiveTrivia(DefineDirectiveTriviaSyntax defineDirective)
        {
            TokenData td = new TokenData(defineDirective.ToString(), defineDirective.GetType().ToString(),
                               defineDirective.Kind(), defineDirective.GetLocation().GetLineSpan().StartLinePosition,
                               defineDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = defineDirective.GetType().ToString();
            td.PpTake = defineDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#undef'
         * directive.
         */
        public override void VisitUndefDirectiveTrivia(UndefDirectiveTriviaSyntax undefDirective)
        {
            TokenData td = new TokenData(undefDirective.ToString(), undefDirective.GetType().ToString(),
                               undefDirective.Kind(), undefDirective.GetLocation().GetLineSpan().StartLinePosition,
                               undefDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = undefDirective.GetType().ToString();
            td.PpTake = undefDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#error'
         * directive.
         */
        public override void VisitErrorDirectiveTrivia(ErrorDirectiveTriviaSyntax errorDirective)
        {
            TokenData td = new TokenData(errorDirective.ToString(), errorDirective.GetType().ToString(),
                               errorDirective.Kind(), errorDirective.GetLocation().GetLineSpan().StartLinePosition,
                               errorDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = errorDirective.GetType().ToString();
            td.PpTake = errorDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#warning'
         * directive.
         */
        public override void VisitWarningDirectiveTrivia(WarningDirectiveTriviaSyntax warningDirective)
        {
            TokenData td = new TokenData(warningDirective.ToString(), warningDirective.GetType().ToString(),
                               warningDirective.Kind(), warningDirective.GetLocation().GetLineSpan().StartLinePosition,
                               warningDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = warningDirective.GetType().ToString();
            td.PpTake = warningDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#line'
         * directive.
         */
        public override void VisitLineDirectiveTrivia(LineDirectiveTriviaSyntax lineDirective)
        {
            TokenData td = new TokenData(lineDirective.ToString(), lineDirective.GetType().ToString(),
                               lineDirective.Kind(), lineDirective.GetLocation().GetLineSpan().StartLinePosition,
                               lineDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = lineDirective.GetType().ToString();
            td.PpTake = lineDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#pragma warning'
         * directive.
         */
        public override void VisitPragmaWarningDirectiveTrivia(PragmaWarningDirectiveTriviaSyntax pragmaWarningDirective)
        {
            TokenData td = new TokenData(pragmaWarningDirective.ToString(), pragmaWarningDirective.GetType().ToString(),
                               pragmaWarningDirective.Kind(), pragmaWarningDirective.GetLocation().GetLineSpan().StartLinePosition,
                               pragmaWarningDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = pragmaWarningDirective.GetType().ToString();
            td.PpTake = pragmaWarningDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for '#pragma checksum'
         * directive.
         */
        public override void VisitPragmaChecksumDirectiveTrivia(PragmaChecksumDirectiveTriviaSyntax pragmaChecksumDirective)
        {
            TokenData td = new TokenData(pragmaChecksumDirective.ToString(), pragmaChecksumDirective.GetType().ToString(),
                               pragmaChecksumDirective.Kind(), pragmaChecksumDirective.GetLocation().GetLineSpan().StartLinePosition,
                               pragmaChecksumDirective.GetLocation().GetLineSpan().EndLinePosition);

            td.ItemType = pragmaChecksumDirective.GetType().ToString();
            td.PpTake = pragmaChecksumDirective.IsActive;

            CountToken(td);
        }

        /**
         * Visitor method for Trivia constructs.
         * Here it is been used for visiting comments.
         */
        public override void VisitTrivia(SyntaxTrivia trivia)
        {

            if (trivia.RawKind >= (int)SyntaxKind.SingleLineCommentTrivia && trivia.RawKind <= (int)SyntaxKind.MultiLineDocumentationCommentTrivia)
            {
                VisitComment(trivia);
                return;
            }
            /* Any section of code like statements, expression, comments etc
             * that are inside a pre-processor directive that evaluates to
             * false are treated as DisabledTextTrivia. Here, they too are
             * marked as inactive.
             */
            else if (trivia.RawKind == (int)SyntaxKind.DisabledTextTrivia)
            {
                markInactiveCode(new TokenData(trivia.GetLocation().GetLineSpan().
                    StartLinePosition, trivia.GetLocation().GetLineSpan().
                    EndLinePosition));
            }
            base.VisitTrivia(trivia);
        }

        #endregion

    }

}